#!/usr/bin/env node
'use strict';

var tslib = require('tslib');
var fs = require('fs');
var commander = require('commander');
var chalk = require('chalk');
var generateCode = require('./generateCode-6c90dbb0.js');
require('path');
require('ts-array-extensions');
require('yaml');
require('json-pointer');
require('util');
require('prettier');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var commander__default = /*#__PURE__*/_interopDefaultLegacy(commander);
var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);

var packageJsonContent = fs__default["default"].readFileSync("".concat(__dirname, "/../package.json"), 'utf-8');
var packageJson = JSON.parse(packageJsonContent);
var version = packageJson.version;
var cmd = new commander__default["default"].Command('generate')
    .description('Generates typescript code from the spec.')
    .arguments('<spec-filename-or-url>')
    .option('--template <template>', 'the type to output. client, server or stubs. Default client')
    .option('--filename <filename>', 'the file to output. Defaults to <service-name><output-type>.generated.ts')
    .option('--service-name <name>', 'the name of the service. Defaults to info.title from the open api document')
    .option('--not-required-type <nullable|optional|both>', 'what type fields should be if they are not in the required list. Defaults to undefined', 'optional')
    .action(function (specFilenameOrUrl, options) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    var logger, output, e_1;
    var _a;
    return tslib.__generator(this, function (_b) {
        switch (_b.label) {
            case 0:
                logger = generateCode.processLogger();
                if (options.template &&
                    !['client', 'server', 'stubs'].includes(options.template.toLowerCase())) {
                    commander__default["default"].help({ error: true });
                }
                if (!['nullable', 'optional', 'both'].includes(options.notRequiredType)) {
                    commander__default["default"].help({ error: true });
                }
                output = (options.template
                    ? options.template.toUpperCase()
                    : 'CLIENT');
                _b.label = 1;
            case 1:
                _b.trys.push([1, 3, , 4]);
                return [4 /*yield*/, generateCode.generateCode({
                        openApiDocumentFilenameOrUrl: specFilenameOrUrl,
                        output: output,
                        outputFilename: options.filename,
                        serviceName: options.serviceName,
                        logger: logger,
                        nonRequiredType: options.notRequiredType
                    })];
            case 2:
                _b.sent();
                logger.log(chalk__default["default"].green('\nCompleted successfully!\n\n'));
                process.exit(0);
                return [3 /*break*/, 4];
            case 3:
                e_1 = _b.sent();
                logger.error("\nFAILED. ".concat(e_1 instanceof Error
                    ? "".concat(e_1.message, "\n").concat((_a = e_1.stack) !== null && _a !== void 0 ? _a : '')
                    : 'Unknown error', "\n\n"));
                process.exit(1);
                return [3 /*break*/, 4];
            case 4: return [2 /*return*/];
        }
    });
}); });
commander__default["default"].program.version(version).addCommand(cmd).parse();
