'use strict';

var tslib = require('tslib');
var path = require('path');
require('ts-array-extensions');
var chalk = require('chalk');
var fs = require('fs');
var yaml = require('yaml');
var jsonPointer = require('json-pointer');
var util = require('util');
var prettier = require('prettier');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
var chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
var yaml__default = /*#__PURE__*/_interopDefaultLegacy(yaml);
var jsonPointer__default = /*#__PURE__*/_interopDefaultLegacy(jsonPointer);
var prettier__default = /*#__PURE__*/_interopDefaultLegacy(prettier);

/* eslint-disable no-console */
var o = process.stdout;
var e = process.stdout;
var progressLogger = function (start) {
    var last = 0;
    return function (m) {
        o.cursorTo(start);
        o.write(m);
        if (m.length < last) {
            o.write(' '.repeat(last - m.length));
        }
        last = m.length;
    };
};
var success = function () { return chalk__default["default"].green(' -> SUCCESS'); };
var warn = function (message) {
    return chalk__default["default"].yellow(" -> WARNING: ".concat(message));
};
var progress = function (message) {
    return chalk__default["default"].blue(" -> ".concat(message));
};
var processLogger = function () {
    var log = function (m) {
        o.write("\n".concat(m));
    };
    return {
        error: function (m) {
            e.write("\n".concat(chalk__default["default"].red(m)));
        },
        log: log,
        create: function (m) {
            log(m);
            var l = progressLogger(m.length);
            l(progress('...'));
            return l;
        }
    };
};

var mapProperties$1 = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        var property = properties[key];
        if ('format' in property && property.format === 'binary') {
            return;
        }
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema$1(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema$1 = function (schema) {
    if ('$ref' in schema) {
        return {
            $ref: schema.$ref.replace('components/schemas', 'definitions')
        };
    }
    return tslib.__assign(tslib.__assign(tslib.__assign(tslib.__assign(tslib.__assign(tslib.__assign({}, schema), ('properties' in schema && schema.properties
        ? { properties: mapProperties$1(schema.properties) }
        : undefined)), (schema.allOf
        ? { allOf: schema.allOf.map(convertSchemaToJsonSchema$1) }
        : undefined)), (schema.anyOf
        ? { anyOf: schema.anyOf.map(convertSchemaToJsonSchema$1) }
        : undefined)), (schema.oneOf
        ? { oneOf: schema.oneOf.map(convertSchemaToJsonSchema$1) }
        : undefined)), ('items' in schema && schema.items
        ? { items: convertSchemaToJsonSchema$1(schema.items) }
        : undefined));
};

// eslint-disable-next-line @typescript-eslint/ban-types
var dereference$2 = function (references, $ref, prefix) {
    var current = {
        $ref: $ref
    };
    while (current && '$ref' in current) {
        if (!current.$ref.startsWith(prefix)) {
            throw new Error("Invalid $ref ".concat($ref));
        }
        var name = current.$ref.substring(current.$ref.lastIndexOf('/') + 1);
        current = references === null || references === void 0 ? void 0 : references[name];
    }
    if (!current)
        throw new Error("Failed to deference ".concat($ref));
    return current;
};

var dereferenceParameter$1 = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$2((_a = doc.components) === null || _a === void 0 ? void 0 : _a.parameters, $ref, '#/components/parameters/');
};

var dereferenceRequestBodyObject = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$2((_a = doc.components) === null || _a === void 0 ? void 0 : _a.requestBodies, $ref, '#/components/requestBodies/');
};

var dereferenceResponseObject$1 = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$2((_a = doc.components) === null || _a === void 0 ? void 0 : _a.responses, $ref, '#/components/responses/');
};

var dereferenceSchemaObject = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$2((_a = doc.components) === null || _a === void 0 ? void 0 : _a.schemas, $ref, '#/components/schemas/');
};

var dereferenceSecurityScheme = function (reference, doc) {
    var _a;
    var $ref = reference.$ref;
    return dereference$2((_a = doc.components) === null || _a === void 0 ? void 0 : _a.securitySchemes, $ref, '#/components/securitySchemes/');
};

var isObjectSchema$1 = function (o) {
    return o.type === 'object' || 'properties' in o;
};

function isReferenceObject$1(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var initUpper = function (value) {
    return value.replace(/(?:^|\s)[a-z]/g, function (m) { return m.toUpperCase(); });
};

var inferOperationId = function (method, path) {
    var _a;
    var varsRegex = /\{.*?\}/gm;
    var vars = (_a = path
        .match(varsRegex)) === null || _a === void 0 ? void 0 : _a.map(function (m) { return m.substring(1, m.length - 1); }).map(initUpper);
    var suffix = (vars === null || vars === void 0 ? void 0 : vars.any()) ? "By".concat(vars.join('And')) : '';
    var remainingPath = path
        .replace(varsRegex, '')
        .replace(/\/|-|_/g, ' ')
        .split(' ')
        .map(initUpper)
        .join('');
    return "".concat(method).concat(remainingPath).concat(suffix);
};

var getSecurityParams = function (document, location) {
    if (!location.security) {
        return [];
    }
    return location.security.flatMap(function (s) {
        return Object.keys(s).map(function (key) {
            var scheme = dereferenceSecurityScheme({ $ref: "#/components/securitySchemes/".concat(key) }, document);
            if (scheme.type !== 'http') {
                throw new Error("Security scheme type ".concat(scheme.type, " not implemented"));
            }
            if (scheme.scheme !== 'bearer') {
                throw new Error("Security scheme ".concat(scheme.scheme, " not implemented"));
            }
            return {
                name: 'authorization',
                in: 'header'
            };
        });
    });
};

var methodsOfPathItemObject$1 = function (pathItemObject) {
    return Object.keys(pathItemObject).compactMap(function (method) {
        if (method === '$ref' ||
            method === 'parameters' ||
            method === 'summary' ||
            method === 'description' ||
            method === 'servers') {
            return undefined;
        }
        var op = pathItemObject[method];
        if (!op) {
            return undefined;
        }
        return [method, op];
    });
};

var mapOperations$1 = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        var _a;
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        var pathParameters = (_a = pathObject.parameters) !== null && _a !== void 0 ? _a : [];
        var globalSecurityParameters = getSecurityParams(document, document);
        return methodsOfPathItemObject$1(pathObject).map(function (_a) {
            var _b = tslib.__read(_a, 2), method = _b[0], operationObject = _b[1];
            var _c = operationObject.operationId, operationId = _c === void 0 ? inferOperationId(method, path) : _c, _d = operationObject.parameters, parameters = _d === void 0 ? [] : _d, responses = operationObject.responses, requestBody = operationObject.requestBody;
            var inlinedParameters = pathParameters
                .concat(parameters)
                .map(function (p) {
                return isReferenceObject$1(p) ? dereferenceParameter$1(p, document) : p;
            })
                .concat(globalSecurityParameters)
                .concat(getSecurityParams(document, operationObject))
                .groupBy(function (p) { return p.name; })
                .compactMap(function (g) { return g.values.first(); });
            var inlinedRequestBody = requestBody && isReferenceObject$1(requestBody)
                ? dereferenceRequestBodyObject(requestBody, document)
                : requestBody;
            var inlinedResponses = Object.keys(responses).compactMap(function (statusCode) {
                var response = responses === null || responses === void 0 ? void 0 : responses[statusCode];
                if (!response)
                    return undefined;
                return {
                    statusCode: statusCode.toString(),
                    response: isReferenceObject$1(response)
                        ? dereferenceResponseObject$1(response, document)
                        : response
                };
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlinedParameters !== null && inlinedParameters !== void 0 ? inlinedParameters : [],
                requestBody: inlinedRequestBody,
                responses: inlinedResponses
            };
        });
    });
};

var typeDefForEnum$2 = function (members) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"".concat(e, "\"");
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'number') {
            return "".concat(e);
        }
        if (typeof e === 'boolean') {
            return e ? 'true' : 'false';
        }
        if (typeof e === 'object') {
            return JSON.stringify(e);
        }
        throw new Error("Unexpected enum member");
    })
        .join(' | ');
};

var reserved = ['protected'];
var aliasIfReserved = function (name) {
    if (reserved.includes(name)) {
        return "".concat(name, ": _").concat(name);
    }
    return name;
};
var aliasNameIfReserved = function (name) {
    if (reserved.includes(name)) {
        return "_".concat(name);
    }
    return name;
};

var initCap$1 = function (src) {
    return src.replace(/\s+.{1}/g, function (v) { return v.slice(-1).toUpperCase(); });
};

var removeIllegalChars = function (src) {
    return src.replace(/[|&;$%@"<>(){}*+,.[\]\-/']/g, ' ');
};

var safeName = function (name) {
    return initCap$1(removeIllegalChars(name)).trim();
};

var fieldTemplate = function (name, required, type, options) {
    var fieldName = safeName(name);
    if (required) {
        return "".concat(fieldName, ": ").concat(type);
    }
    var nonRequiredType = options.nonRequiredType;
    var fieldNameWithOptional = ['optional', 'both'].includes(nonRequiredType)
        ? "\"".concat(fieldName, "\"?")
        : "\"".concat(fieldName, "\"");
    var typeWithNull = ['nullable', 'both'].includes(nonRequiredType)
        ? "".concat(type, " | null")
        : type;
    return "".concat(fieldNameWithOptional, ": ").concat(typeWithNull);
};

var objectTemplate = function (fields, options) {
    return fields.none()
        ? 'unknown'
        : "{".concat(fields
            .map(function (f) { return fieldTemplate(f.name, f.required, f.type, options); })
            .join(';'), "}");
};

var clientMethodTemplate = function (_a, options) {
    var httpMethod = _a.httpMethod, methodName = _a.methodName, openApiPath = _a.openApiPath, pathParams = _a.pathParams, queryArrayFormat = _a.queryArrayFormat, queryParams = _a.queryParams, headerParams = _a.headerParams, body = _a.body, responses = _a.responses;
    var bodyArg = (function () {
        if (!body)
            return undefined;
        var typeDef = body.type === 'json'
            ? body.jsonType
            : objectTemplate(body.fields.map(function (f) { return ({
                name: f.name,
                required: f.required,
                type: f.binary ? 'Blob' : 'string'
            }); }), options);
        return {
            name: 'body',
            required: body.required,
            type: typeDef
        };
    })();
    var functionArguments = queryParams
        .concat(headerParams)
        .concat(pathParams.map(function (_a) {
        var name = _a.name, type = _a.type;
        return ({ name: name, required: true, type: type });
    }))
        .concat(bodyArg ? [bodyArg] : []);
    var functionArgumentSignature = (functionArguments.any()
        ? ["args: ".concat(objectTemplate(functionArguments, options))]
        : [])
        .concat('options?: RequestInit')
        .join(',\n');
    var responseType = responses
        .map(function (r) {
        var statusCodeType = r.statusCode === 'default' ? 'number' : r.statusCode;
        var dataType = (function () {
            if (r.type === 'binary')
                return 'globalThis.Blob';
            if (r.type === 'text')
                return 'string';
            if (r.type === 'json')
                return r.jsonType;
            return 'undefined';
        })();
        return "ResponseWithData<".concat(statusCodeType, ", ").concat(dataType, ">");
    })
        .join(' | ');
    var paramNames = headerParams
        .concat(queryParams)
        .concat(pathParams.map(function (_a) {
        var name = _a.name, type = _a.type;
        return ({ name: name, required: true, type: type });
    }))
        .map(function (_a) {
        var name = _a.name;
        return name;
    })
        .concat(bodyArg ? [bodyArg.name] : []);
    var decomposeParameters = paramNames.any()
        ? "const {".concat(paramNames
            .map(safeName)
            .map(aliasIfReserved)
            .join(', '), "} = args;")
        : '';
    var composeQuery = queryParams.any()
        ? "const query = qs.stringify({ ".concat(queryParams
            .map(function (q) { return "[\"".concat(q.name, "\"]: ").concat(safeName(aliasIfReserved(q.name))); })
            .join(', '), " }").concat(queryArrayFormat === 'comma' ? ", { arrayFormat: 'comma' }" : '', ");")
        : '';
    var composeFormData = (body === null || body === void 0 ? void 0 : body.type) === 'form'
        ? "const formData = new FormData();\n      ".concat(body.fields
            .map(function (f) {
            return "if (body.".concat(f.name, ") formData.append('").concat(f.name, "', body.").concat(f.name, ");");
        })
            .join('\n'), "\n    ")
        : '';
    var url = "`${this.baseUrl}".concat(openApiPath.replace(/\{(?:.*?)\}/g, function (x) { return "$".concat(x); })).concat(queryParams.any() ? '?${query}' : '', "`");
    var headers = headerParams
        .map(function (h) {
        if (h.required) {
            return "[\"".concat(h.name, "\"]: ").concat(safeName(aliasIfReserved(h.name)));
        }
        return "...(typeof ".concat(safeName(aliasNameIfReserved(h.name)), " !== 'undefined' && ").concat(safeName(aliasNameIfReserved(h.name)), " !== null ? { [\"").concat(h.name, "\"]: ").concat(safeName(aliasIfReserved(h.name)), " } : {})");
    })
        .concat((body === null || body === void 0 ? void 0 : body.type) === 'json' ? ["'Content-Type': 'application/json'"] : []);
    var fetchOpts = "{\n      ".concat(tslib.__spreadArray(tslib.__spreadArray(tslib.__spreadArray(tslib.__spreadArray([
        'method'
    ], tslib.__read(((body === null || body === void 0 ? void 0 : body.type) === 'json' ? ['body: JSON.stringify(body)'] : [])), false), tslib.__read(((body === null || body === void 0 ? void 0 : body.type) === 'form' ? ['body: formData'] : [])), false), tslib.__read((headers.any() ? ["headers: { ".concat(headers.join(',\n'), " }")] : [])), false), [
        '...options'
    ], false).join(',\n'), "\n    }");
    var responseSwitch = "\n    switch ($status) {\n      ".concat(responses
        .map(function (r) {
        var bodyExpression = (function () {
            if (r.type === 'json') {
                return "(await response.json()) as ".concat(r.jsonType);
            }
            if (r.type === 'binary')
                return 'await response.blob()';
            if (r.type === 'text')
                return 'await response.text()';
            return 'undefined';
        })();
        return "\n              ".concat(r.statusCode === 'default' ? "default" : "case ".concat(r.statusCode), ": return {\n                status: $status,\n                body: ").concat(bodyExpression, "\n              }\n            ");
    })
        .join('\n'), "\n\n      ").concat(responses.every(function (r) { return r.statusCode !== 'default'; })
        ? "\n          default:\n            throw new Error(" +
            '`Unexpected status ${$status}`' +
            ")\n        "
        : '', "\n    }\n  ");
    return "public async \"".concat(methodName, "\"(").concat(functionArgumentSignature, "): \n    Promise<").concat(responseType, "> {\n\n    ").concat(decomposeParameters, "\n\n    ").concat(composeQuery, "\n\n    ").concat(composeFormData, "\n\n    const method = '").concat(httpMethod.toUpperCase(), "',\n    const url = ").concat(url, ";\n\n    const response = await fetch(url, ").concat(fetchOpts, ");\n\n    const { status: $status } = response;\n\n    this.logger?.('REST API Call', { method, url, status: $status });\n\n    ").concat(responseSwitch, "\n  }");
};

var parametersTypeNameTemplate = function (operationId, iin) {
    return "".concat(initUpper(safeName(operationId)), "Request").concat(initUpper(iin === 'formData' ? 'body' : iin));
};

var initLower = function (value) {
    return value.replace(/(?:^|\s)[A-Z]/g, function (m) { return m.toLowerCase(); });
};

var responseBodyTypeNameTemplate = function (operationId, statusCode) {
    return "".concat(initUpper(safeName(operationId))).concat(initUpper(statusCode), "ResponseBody");
};

var serviceNameTemplate = function (title) {
    return initCap$1(removeIllegalChars(title)).trim();
};

var typeDefForObject$2 = function (objectSchema, recursiveLookup, options, emitType) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject, options, emitType);
        return fieldTemplate(propertyName, isRequired, propertyType, options);
    });
    return "{ ".concat(propertyList.join('; '), " }");
};

/* eslint-disable no-restricted-syntax */
/* eslint-disable no-continue */
var initCap = function (value) {
    return value
        .toLowerCase()
        .replace(/(?:^|\s)[a-z]/g, function (m) { return m.toUpperCase(); });
};
function getCharClass(char) {
    if (/[A-Z]/.test(char))
        return 'UPPER';
    if (/[a-z]/.test(char))
        return 'LOWER';
    if (/[0-9]/.test(char))
        return 'NUMBER';
    return 'OTHER';
}
function getWords(input) {
    var currentWord, input_1, input_1_1, char, charClass, lastChar, e_1_1;
    var e_1, _a;
    return tslib.__generator(this, function (_b) {
        switch (_b.label) {
            case 0:
                currentWord = [];
                _b.label = 1;
            case 1:
                _b.trys.push([1, 9, 10, 11]);
                input_1 = tslib.__values(input), input_1_1 = input_1.next();
                _b.label = 2;
            case 2:
                if (!!input_1_1.done) return [3 /*break*/, 8];
                char = input_1_1.value;
                charClass = getCharClass(char);
                lastChar = currentWord.last();
                // start of new word
                if (!lastChar) {
                    if (charClass !== 'OTHER') {
                        currentWord.push(char);
                    }
                    return [3 /*break*/, 7];
                }
                if (!(charClass === 'UPPER' && getCharClass(lastChar) !== 'UPPER')) return [3 /*break*/, 4];
                return [4 /*yield*/, currentWord.join('')];
            case 3:
                _b.sent();
                currentWord = [char];
                return [3 /*break*/, 7];
            case 4:
                if (!(charClass === 'OTHER' && currentWord.any())) return [3 /*break*/, 6];
                return [4 /*yield*/, currentWord.join('')];
            case 5:
                _b.sent();
                currentWord = [];
                return [3 /*break*/, 7];
            case 6:
                currentWord.push(char);
                _b.label = 7;
            case 7:
                input_1_1 = input_1.next();
                return [3 /*break*/, 2];
            case 8: return [3 /*break*/, 11];
            case 9:
                e_1_1 = _b.sent();
                e_1 = { error: e_1_1 };
                return [3 /*break*/, 11];
            case 10:
                try {
                    if (input_1_1 && !input_1_1.done && (_a = input_1.return)) _a.call(input_1);
                }
                finally { if (e_1) throw e_1.error; }
                return [7 /*endfinally*/];
            case 11:
                if (!currentWord.any()) return [3 /*break*/, 13];
                return [4 /*yield*/, currentWord.join('')];
            case 12:
                _b.sent();
                _b.label = 13;
            case 13: return [2 /*return*/];
        }
    });
}
var typeName = function (schemaName) {
    return tslib.__spreadArray([], tslib.__read(getWords(schemaName)), false).map(initCap).join('');
};

var typeDefForReference$1 = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return typeName(name);
};

var typeDefForSchema$2 = function (schema, options, emitType) {
    if ('$ref' in schema) {
        return typeDefForReference$1(schema.$ref);
    }
    var _a = options.binaryType, binaryType = _a === void 0 ? 'Buffer' : _a;
    var enumProp = schema.enum, allOf = schema.allOf, oneOf = schema.oneOf, anyOf = schema.anyOf, title = schema.title;
    var preliminaryType = (function () {
        if (enumProp) {
            return typeDefForEnum$2(enumProp);
        }
        var base = isObjectSchema$1(schema)
            ? typeDefForObject$2(schema, typeDefForSchema$2, options, emitType)
            : null;
        if (allOf) {
            return allOf
                .map(function (m) { return typeDefForSchema$2(m, options, emitType); })
                .concat(base ? [base] : [])
                .join(' & ');
        }
        var unionOf = (oneOf !== null && oneOf !== void 0 ? oneOf : []).concat(anyOf !== null && anyOf !== void 0 ? anyOf : []);
        if (unionOf.length > 0) {
            var union = unionOf
                .map(function (m) { return typeDefForSchema$2(m, options, emitType); })
                .join(' | ');
            return base ? "".concat(base, " & (").concat(union, ")") : union;
        }
        if (!schema.type) {
            // default to object
            return base !== null && base !== void 0 ? base : 'unknown';
        }
        if (schema.type === 'object') {
            return base !== null && base !== void 0 ? base : 'unknown';
        }
        if (schema.type === 'string') {
            if (schema.format === 'binary') {
                return binaryType;
            }
            return 'string';
        }
        if (schema.type === 'boolean' || schema.type === 'number') {
            return schema.type;
        }
        if (schema.type === 'integer') {
            return 'number';
        }
        if (schema.type === 'array') {
            var items = schema.items;
            if (!items) {
                throw new Error("Array nodes must have an items property.");
            }
            if (Array.isArray(items)) {
                throw new Error("Array items property must be singular.");
            }
            return "Array<".concat(typeDefForSchema$2(items, options, emitType), ">");
        }
        return 'unknown';
    })();
    var code = 'nullable' in schema && schema.nullable
        ? "".concat(preliminaryType, " | null")
        : preliminaryType;
    if (title) {
        var type = typeName(title);
        emitType(type, code);
        return type;
    }
    return code;
};

var fromV3$4 = function (document, options, emitType, log) {
    return mapOperations$1(document).map(function (_a) {
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId, responses = _a.responses, requestBody = _a.requestBody;
        log === null || log === void 0 ? void 0 : log(progress("Adding ".concat(method, " ").concat(path)));
        var body = (function () {
            var _a;
            if (!requestBody)
                return null;
            if ('application/json' in requestBody.content) {
                var schema = requestBody.content['application/json'].schema;
                return {
                    type: 'json',
                    required: (_a = requestBody.required) !== null && _a !== void 0 ? _a : false,
                    jsonType: schema
                        ? typeDefForSchema$2(schema, options, emitType)
                        : 'unknown'
                };
            }
            if ('multipart/form-data' in requestBody.content) {
                var schema = requestBody.content['multipart/form-data'].schema;
                if (!schema)
                    return null;
                var inlinedSchema_1 = isReferenceObject$1(schema)
                    ? dereferenceSchemaObject(schema, document)
                    : schema;
                var fields = inlinedSchema_1.properties
                    ? Object.keys(inlinedSchema_1.properties).map(function (k) {
                        var _a, _b, _c;
                        var property = (_a = inlinedSchema_1.properties) === null || _a === void 0 ? void 0 : _a[k];
                        var inlinedProperty = isReferenceObject$1(property)
                            ? dereferenceSchemaObject(property, document)
                            : property;
                        return {
                            name: k.toString(),
                            required: (_c = (_b = inlinedSchema_1.required) === null || _b === void 0 ? void 0 : _b.includes(k.toString())) !== null && _c !== void 0 ? _c : false,
                            binary: (inlinedProperty === null || inlinedProperty === void 0 ? void 0 : inlinedProperty.format) === 'binary'
                        };
                    })
                    : [];
                return {
                    type: 'form',
                    required: requestBody.required,
                    fields: fields
                };
            }
            return null;
        })();
        return clientMethodTemplate({
            httpMethod: method,
            methodName: operationId,
            openApiPath: path,
            pathParams: parameters
                .filter(function (p) { return p.in === 'path'; })
                .map(function (p) { return ({
                name: p.name,
                type: p.schema
                    ? typeDefForSchema$2(p.schema, options, emitType)
                    : 'string'
            }); }),
            queryArrayFormat: parameters
                .filter(function (p) { return p.in === 'query'; })
                .some(function (p) { return p.explode === false; })
                ? 'comma'
                : 'repeat',
            queryParams: parameters
                .filter(function (p) { return p.in === 'query'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false,
                    type: p.schema
                        ? typeDefForSchema$2(p.schema, options, emitType)
                        : 'string'
                });
            }),
            headerParams: parameters
                .filter(function (p) { return p.in === 'header'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false,
                    type: p.schema
                        ? typeDefForSchema$2(p.schema, options, emitType)
                        : 'string'
                });
            }),
            body: body,
            responses: responses.map(function (_a) {
                var statusCode = _a.statusCode, response = _a.response;
                if (!response.content ||
                    Object.keys(response.content).length === 0) {
                    return {
                        statusCode: statusCode,
                        type: 'none'
                    };
                }
                if (response.content) {
                    if ('application/json' in response.content) {
                        var schema = response.content['application/json'].schema;
                        return {
                            statusCode: statusCode,
                            type: 'json',
                            jsonType: schema
                                ? typeDefForSchema$2(schema, options, emitType)
                                : 'unknown'
                        };
                    }
                    if ('text/html' in response.content) {
                        return {
                            statusCode: statusCode,
                            type: 'text'
                        };
                    }
                }
                return {
                    statusCode: statusCode,
                    type: 'binary'
                };
            })
        }, options);
    });
};

var mapProperties = function (properties) {
    var r = {};
    Object.keys(properties).forEach(function (key) {
        // eslint-disable-next-line @typescript-eslint/no-use-before-define
        r[key] = convertSchemaToJsonSchema(properties[key]);
    });
    return r;
};
var convertSchemaToJsonSchema = function (schemas) {
    /**
     * This currently does nothing, but is a placeholder for removing
     * unsupported features from Open Api
     */
    var convert = function (schema) {
        var properties = schema.properties, rest = tslib.__rest(schema, ["properties"]);
        return tslib.__assign(tslib.__assign({}, rest), (properties ? { properties: mapProperties(properties) } : undefined));
    };
    if (Array.isArray(schemas)) {
        return { anyOf: schemas.map(convert) };
    }
    return convert(schemas);
};

var dereference$1 = function (references, $ref, prefix) {
    if (!$ref.startsWith(prefix)) {
        throw new Error("Invalid $ref ".concat($ref));
    }
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    var maybe = references === null || references === void 0 ? void 0 : references[name];
    if (!maybe)
        throw new Error("Failed to deference ".concat($ref));
    return maybe;
};

var dereferenceParameter = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference$1(doc.parameters, $ref, '#/parameters/');
};

var dereferenceResponseObject = function (reference, doc) {
    var $ref = reference.$ref;
    return dereference$1(doc.responses, $ref, '#/responses/');
};

function isReferenceObject(object) {
    if (typeof object !== 'object') {
        return false;
    }
    if (object === null) {
        return false;
    }
    return '$ref' in object;
}

var methodsOfPathItemObject = function (pathItemObject) {
    return Object.keys(pathItemObject).compactMap(function (method) {
        if (method === '$ref' || method === 'parameters') {
            return undefined;
        }
        var op = pathItemObject[method];
        if (!op) {
            return undefined;
        }
        return [method, op];
    });
};

var mapOperations = function (document) {
    return Object.keys(document.paths).flatMap(function (path) {
        var _a;
        if (typeof path !== 'string')
            return [];
        var pathObject = document.paths[path];
        var pathParameters = (_a = pathObject.parameters) !== null && _a !== void 0 ? _a : [];
        return methodsOfPathItemObject(pathObject).map(function (_a) {
            var _b = tslib.__read(_a, 2), method = _b[0], operationObject = _b[1];
            var _c = operationObject.operationId, operationId = _c === void 0 ? inferOperationId(method, path) : _c, _d = operationObject.parameters, parameters = _d === void 0 ? [] : _d, responses = operationObject.responses;
            var inlinedParameters = pathParameters
                .concat(parameters)
                .map(function (p) {
                return isReferenceObject(p) ? dereferenceParameter(p, document) : p;
            });
            var inlinedResponses = Object.keys(responses).compactMap(function (statusCode) {
                var response = responses[statusCode];
                if (!response)
                    return undefined;
                return {
                    statusCode: statusCode.toString(),
                    response: isReferenceObject(response)
                        ? dereferenceResponseObject(response, document)
                        : response
                };
            });
            return {
                path: path,
                method: method,
                operationId: operationId,
                operation: operationObject,
                parameters: inlinedParameters !== null && inlinedParameters !== void 0 ? inlinedParameters : [],
                responses: inlinedResponses
            };
        });
    });
};

var typeDefForEnum$1 = function (members) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"".concat(e, "\"");
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'number') {
            return "".concat(e);
        }
        if (typeof e === 'boolean') {
            return e ? 'true' : 'false';
        }
        if (typeof e === 'object') {
            return JSON.stringify(e);
        }
        throw new Error("Unexpected enum member");
    })
        .join(' | ');
};

var typeDefForObject$1 = function (objectSchema, recursiveLookup, options, emitType) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject, options, emitType);
        return fieldTemplate(propertyName, isRequired, propertyType, options);
    });
    return "{ ".concat(propertyList.join('; '), " }");
};

var typeDefForSchema$1 = function (schemas, options, emitType) {
    var gen = function (schema) {
        if (typeof schema === 'boolean') {
            throw new Error("Schema nodes must be an object.");
        }
        if ('$ref' in schema && schema.$ref) {
            return typeDefForReference$1(schema.$ref);
        }
        var _a = schema.type, type = _a === void 0 ? 'object' : _a, enumProp = schema.enum, allOf = schema.allOf, title = schema.title;
        var code = (function () {
            if (enumProp) {
                return typeDefForEnum$1(enumProp);
            }
            if (type === 'object') {
                var def = typeDefForObject$1(schema, typeDefForSchema$1, options, emitType);
                if (allOf) {
                    return allOf
                        .map(function (m) {
                        return typeDefForSchema$1(m, options, emitType);
                    })
                        .concat(def)
                        .join(' & ');
                }
                return def;
            }
            if (allOf) {
                return allOf
                    .map(function (m) {
                    return typeDefForSchema$1(m, options, emitType);
                })
                    .join(' & ');
            }
            if (type === 'boolean' ||
                type === 'null' ||
                type === 'number' ||
                type === 'string') {
                return type;
            }
            if (type === 'integer') {
                return 'number';
            }
            if (type === 'array') {
                var items = schema.items;
                if (!items) {
                    throw new Error("Array nodes must have an items property.");
                }
                if (Array.isArray(items)) {
                    throw new Error("Array items property must be singular.");
                }
                return "Array<".concat(typeDefForSchema$1(items, options, emitType), ">");
            }
            return 'unknown';
        })();
        if (title) {
            var typeTitle = typeName(title);
            emitType(typeTitle, code);
            return typeTitle;
        }
        return code;
    };
    if (Array.isArray(schemas)) {
        return schemas.map(gen).join(' | ');
    }
    return gen(schemas);
};

var fromV2$4 = function (document, options, emitType, log) {
    return mapOperations(document).map(function (_a) {
        var _b;
        var method = _a.method, parameters = _a.parameters, path = _a.path, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding ".concat(method, " ").concat(path)));
        var jsonBody = function () {
            var bodyArg = parameters.filter(function (p) { return p.in === 'body'; }).first();
            if (!bodyArg || bodyArg.in !== 'body')
                return null;
            var schema = bodyArg.schema, _a = bodyArg.required, required = _a === void 0 ? false : _a;
            return {
                type: 'json',
                jsonType: typeDefForSchema$1(schema, options, emitType),
                required: required
            };
        };
        var formDataBody = function () {
            var bodyArg = parameters.filter(function (p) { return p.in === 'formData'; });
            if (bodyArg.length === 0)
                return null;
            return {
                type: 'form',
                fields: bodyArg.map(function (p) {
                    var _a;
                    return ({
                        name: p.name,
                        required: (_a = p.required) !== null && _a !== void 0 ? _a : false,
                        binary: p.type === 'file'
                    });
                }),
                required: true
            };
        };
        return clientMethodTemplate({
            httpMethod: method,
            methodName: operationId,
            openApiPath: path,
            pathParams: parameters
                .filter(function (p) { return p.in === 'path'; })
                .map(function (p) { return ({
                name: p.name,
                type: typeDefForSchema$1(tslib.__assign(tslib.__assign({}, p), { required: [] }), options, emitType)
            }); }),
            queryArrayFormat: parameters
                .filter(function (p) { return p.in === 'query'; })
                .some(function (p) { return p.collectionFormat === 'csv'; })
                ? 'comma'
                : 'repeat',
            queryParams: parameters
                .filter(function (p) { return p.in === 'query'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false,
                    type: p.schema
                        ? typeDefForSchema$1(p.schema, options, emitType)
                        : 'string'
                });
            }),
            headerParams: parameters
                .filter(function (p) { return p.in === 'header'; })
                .map(function (p) {
                var _a;
                return ({
                    name: p.name,
                    required: (_a = p.required) !== null && _a !== void 0 ? _a : false,
                    type: p.schema
                        ? typeDefForSchema$1(p.schema, options, emitType)
                        : 'string'
                });
            }),
            body: (_b = jsonBody()) !== null && _b !== void 0 ? _b : formDataBody(),
            responses: responses.map(function (_a) {
                var _b;
                var statusCode = _a.statusCode, response = _a.response;
                if (((_b = response.schema) === null || _b === void 0 ? void 0 : _b.type) === 'file') {
                    return {
                        statusCode: statusCode,
                        type: 'binary'
                    };
                }
                var schema = response.schema;
                if (schema) {
                    return {
                        statusCode: statusCode,
                        type: 'json',
                        jsonType: schema
                            ? typeDefForSchema$1(schema, options, emitType)
                            : 'unknown'
                    };
                }
                return {
                    statusCode: statusCode,
                    type: 'none'
                };
            })
        }, options);
    });
};

var generateMethods = function (openApiDocument, options, emitType, log) {
    return 'swagger' in openApiDocument
        ? fromV2$4(openApiDocument, options, emitType, log)
        : fromV3$4(openApiDocument, options, emitType, log);
};

var generateClient = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName, nonRequiredType = _a.nonRequiredType;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating client');
    var types = {};
    var methods = generateMethods(openApiDocument, { nonRequiredType: nonRequiredType }, function (name, def) {
        types[name] = def;
    }, log);
    var code = "\n    import qs from 'qs';\n    import { fetch } from 'cross-fetch';\n\n    ".concat(Object.entries(types)
        .map(function (_a) {
        var _b = tslib.__read(_a, 2), name = _b[0], def = _b[1];
        return "export type ".concat(name, " = ").concat(def, ";");
    })
        .join('\n\n'), "\n\n    type ResponseWithData<TStatus, TData> = {\n      status: TStatus;\n      body: TData;\n    };\n      \n    type LogFn = (message: string, data: Record<string, unknown>) => void;\n\n    export class ").concat(serviceName, "Client {\n      public constructor(\n        public readonly baseUrl: string,\n        private readonly logger?: LogFn\n      ) {}\n      \n      ").concat(methods.join('\n\n'), "\n    }\n  ");
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var controllerMiddlewareTemplate = function (handlers) {
    return "\n    export interface RequestResolver {\n      resolve<T>(token: \n        | { new (...args: any[]): T; }\n        | string\n        | symbol): T;\n    }\n\n    @injectable()\n    export class RequestResolverFactory {\n      constructor(public readonly forRequest: (req: Request) => RequestResolver) {};\n    }\n\n    @injectable()\n    export class ControllerMiddleware {\n      constructor(\n        private readonly resolver: RequestResolverFactory\n      ) {}\n\n      private static validate = new Validator({ strict: false, coerceTypes: true }).validate;\n\n      ".concat(handlers
        .map(function (h) { return "\n      private add".concat(h.controllerTypeName, "(router: Router): void {\n        router.").concat(h.httpMethod, "(\n          '").concat(h.expressPath, "',\n    \n          ").concat(h.bodySchema || h.querySchema || h.pathSchema
        ? "\n          ControllerMiddleware.validate({\n            ".concat(h.bodySchema ? "body: ".concat(JSON.stringify(h.bodySchema), ",") : '', "\n            ").concat(h.querySchema ? "query: ".concat(JSON.stringify(h.querySchema), ",") : '', "\n            ").concat(h.pathSchema ? "params: ".concat(JSON.stringify(h.pathSchema), ",") : '', "\n          }),\n          ")
        : '', "\n    \n          asyncRequestHandler<").concat(h.requestTypeName, ", ").concat(h.responseTypeName, ">(\n            async (req, res, next) => {\n              const controller = this.resolver\n                .forRequest(req)\n                .resolve<").concat(h.controllerTypeName, ">(\n                  nameof<").concat(h.controllerTypeName, ">()\n                );\n    \n              await controller.").concat(h.controllerMethodName, "(req, res, next);\n            }\n          )\n        );\n      }\n      "); })
        .join('\n'), "\n\n      private static validationErrorMiddleware: ErrorRequestHandler = (error, _, response, next) => {\n        if (error instanceof ValidationError) {\n          response.status(400).send({\n            type: 'REQUEST_VALIDATION_FAILED',\n            fields: error.validationErrors\n          });\n          next();\n        } else {\n          next(error);\n        }\n      };    \n    \n      public apply(expressApp: Express): void {\n        const router = Router();\n\n      ").concat(handlers
        .map(function (h) { return "this.add".concat(h.controllerTypeName, "(router);"); })
        .join('\n'), "\n    \n        expressApp.use(router);\n\n        expressApp.use(ControllerMiddleware.validationErrorMiddleware);\n      }\n    }");
};

var controllerTemplate = function (_a) {
    var requestTypeName = _a.requestTypeName, responseTypeName = _a.responseTypeName, controllerTypeName = _a.controllerTypeName, controllerMethodName = _a.controllerMethodName, _b = _a.pathParamsType, pathParamsType = _b === void 0 ? 'ParamsDictionary' : _b, _c = _a.responseBodyType, responseBodyType = _c === void 0 ? 'unknown' : _c, _d = _a.bodyParamType, bodyParamType = _d === void 0 ? 'unknown' : _d, _e = _a.queryParamsType, queryParamsType = _e === void 0 ? 'ParsedQs' : _e, _f = _a.statusCodeType, statusCodeType = _f === void 0 ? 'number' : _f;
    return "\n    export type ".concat(requestTypeName, " = Request<\n      ").concat(pathParamsType, ",\n      ").concat(responseBodyType, ",\n      ").concat(bodyParamType, ",\n      ").concat(queryParamsType, ",\n      Record<string, any>\n    >;\n\n    export type ").concat(responseTypeName, " = Response<\n      ").concat(responseBodyType, ",\n      Record<string, any>,\n      ").concat(statusCodeType, "\n    >;\n\n    export interface ").concat(controllerTypeName, " {\n      ").concat(controllerMethodName, "(\n        req: ").concat(requestTypeName, ",\n        res: ").concat(responseTypeName, ",\n        next: NextFunction\n      ): Promise<unknown | void>;\n    }");
};

var controllersTemplate = function (controllers) {
    return "\n    ".concat(controllers.map(controllerTemplate).join(';\n'), "\n\n    ").concat(controllerMiddlewareTemplate(controllers), "\n  ");
};

var fromV2$3 = function (document, options, emitType, log) {
    var handlers = mapOperations(document).map(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses, path = _a.path, method = _a.method;
        log === null || log === void 0 ? void 0 : log(progress("Adding ".concat(method, " ").concat(path)));
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            var name = responseBodyTypeNameTemplate(operationId, statusCode);
            emitType(name, typeDefForSchema$1(schema, { nonRequiredType: 'optional' }, emitType));
            return name;
        })
            .join(' | ') || undefined;
        var statusCodeType = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var nameAndSchemaForParams = function (type) {
            var params = parameters.filter(function (p) { return p.in === type; });
            if (params.none()) {
                return [undefined, undefined];
            }
            if (type === 'body') {
                var schemas = params.map(function (p) { return p.schema; });
                var typeName_1 = parametersTypeNameTemplate(operationId, 'body');
                log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName_1)));
                emitType(typeName_1, typeDefForSchema$1(schemas, options, emitType));
                return [
                    typeName_1,
                    convertSchemaToJsonSchema(schemas)
                ];
            }
            var typeName = parametersTypeNameTemplate(operationId, type);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            params.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                if (rest.type === 'file') {
                    return;
                }
                var propertyType = (function () {
                    if (rest.type === 'array') {
                        return { anyOf: [rest, rest.items] };
                    }
                    return rest;
                })();
                schema.properties[name] = propertyType;
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required), false), [name], false);
                }
            });
            emitType(typeName, typeDefForSchema$1(schema, options, emitType));
            return [typeName, convertSchemaToJsonSchema(schema)];
        };
        var _b = tslib.__read(nameAndSchemaForParams('body'), 2), bodyParamType = _b[0], bodySchema = _b[1];
        var _c = tslib.__read(nameAndSchemaForParams('query'), 2), queryParamsType = _c[0], querySchema = _c[1];
        var _d = tslib.__read(nameAndSchemaForParams('path'), 2), pathParamsType = _d[0], pathSchema = _d[1];
        var requestTypeName = "".concat(initUpper(operationId), "Request");
        var responseTypeName = "".concat(initUpper(operationId), "Response");
        var controllerTypeName = "".concat(initUpper(operationId), "Controller");
        var controllerMethodName = initLower(operationId);
        var expressPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":".concat(x.substr(1, x.length - 2)); });
        log === null || log === void 0 ? void 0 : log(progress(operationId));
        return {
            httpMethod: method,
            expressPath: expressPath,
            requestTypeName: requestTypeName,
            responseTypeName: responseTypeName,
            controllerTypeName: controllerTypeName,
            controllerMethodName: controllerMethodName,
            pathParamsType: pathParamsType,
            responseBodyType: responseBodyType,
            bodyParamType: bodyParamType,
            queryParamsType: queryParamsType,
            statusCodeType: statusCodeType,
            bodySchema: bodySchema,
            querySchema: querySchema,
            pathSchema: pathSchema
        };
    });
    return controllersTemplate(handlers);
};

var convertFormDataToSchemaObject$1 = function (formData) {
    if (!formData || !formData.schema)
        return undefined;
    return tslib.__assign({ type: 'object' }, formData.schema);
};
var fromV3$3 = function (document, options, emitType, log) {
    var handlers = mapOperations$1(document).map(function (_a) {
        var path = _a.path, operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses, requestBody = _a.requestBody, method = _a.method;
        var responseBodyType = responses
            .compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            if (!response.content ||
                !('application/json' in response.content)) {
                return undefined;
            }
            var schema = response.content['application/json'].schema;
            var name = responseBodyTypeNameTemplate(operationId, statusCode);
            if (schema) {
                emitType(name, typeDefForSchema$2(schema, options, emitType));
            }
            return name;
        })
            .join(' | ') || undefined;
        var statusCodeType = responses
            .map(function (_a) {
            var statusCode = _a.statusCode;
            return statusCode === 'default' ? 'number' : statusCode;
        })
            .join(' | ');
        var _b = tslib.__read((function () {
            var _a;
            if (!requestBody) {
                return [undefined, undefined];
            }
            var structuredContent = [
                (_a = requestBody.content['application/json']) === null || _a === void 0 ? void 0 : _a.schema,
                convertFormDataToSchemaObject$1(requestBody.content['multipart/form-data'])
            ].compact();
            if (structuredContent.length === 0) {
                return [undefined, undefined];
            }
            var schema = structuredContent.length > 1
                ? {
                    anyOf: structuredContent
                }
                : structuredContent[0];
            var typeName = parametersTypeNameTemplate(operationId, 'body');
            emitType(typeName, typeDefForSchema$2(schema, options, emitType));
            return [typeName, convertSchemaToJsonSchema$1(schema)];
        })(), 2), bodyParamType = _b[0], bodySchema = _b[1];
        var nameAndSchemaForParams = function (type) {
            var params = parameters.filter(function (p) { return p.in === type; });
            if (params.none()) {
                return [undefined, undefined];
            }
            var typeName = parametersTypeNameTemplate(operationId, type);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            params.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                var propertyType = (function () {
                    if (!rest.schema) {
                        return { type: 'string' };
                    }
                    if ('$ref' in rest.schema) {
                        return rest.schema;
                    }
                    if (rest.schema.type === 'array') {
                        return {
                            anyOf: [rest.schema, rest.schema.items]
                        };
                    }
                    return rest.schema;
                })();
                schema.properties[name] = propertyType;
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required), false), [name], false);
                }
            });
            emitType(typeName, typeDefForSchema$2(schema, options, emitType));
            return [typeName, convertSchemaToJsonSchema$1(schema)];
        };
        var _c = tslib.__read(nameAndSchemaForParams('query'), 2), queryParamsType = _c[0], querySchema = _c[1];
        var _d = tslib.__read(nameAndSchemaForParams('path'), 2), pathParamsType = _d[0], pathSchema = _d[1];
        var requestTypeName = "".concat(initUpper(operationId), "Request");
        var responseTypeName = "".concat(initUpper(operationId), "Response");
        var controllerTypeName = "".concat(initUpper(operationId), "Handler");
        var controllerMethodName = initLower(operationId);
        var expressPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":".concat(x.substr(1, x.length - 2)); });
        log === null || log === void 0 ? void 0 : log(progress(operationId));
        return {
            httpMethod: method,
            expressPath: expressPath,
            requestTypeName: requestTypeName,
            responseTypeName: responseTypeName,
            controllerTypeName: controllerTypeName,
            controllerMethodName: controllerMethodName,
            operationId: operationId,
            pathParamsType: pathParamsType,
            pathSchema: pathSchema,
            responseBodyType: responseBodyType,
            bodyParamType: bodyParamType,
            bodySchema: bodySchema,
            queryParamsType: queryParamsType,
            querySchema: querySchema,
            statusCodeType: statusCodeType
        };
    });
    return controllersTemplate(handlers);
};

var generateControllers = function (openApiDocument, options, emitType, log) {
    return 'swagger' in openApiDocument
        ? fromV2$3(openApiDocument, options, emitType, log)
        : fromV3$3(openApiDocument, options, emitType, log);
};

var generateServerClasses = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, nonRequiredType = _a.nonRequiredType;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server');
    var types = {};
    var controllers = generateControllers(openApiDocument, { nonRequiredType: nonRequiredType }, function (name, def) {
        types[name] = def;
    }, log);
    var code = "\n    import { Request, Response, Express, NextFunction, ParamsDictionary, ErrorRequestHandler } from 'express-serve-static-core';\n    import { Router } from 'express';\n    import { asyncRequestHandler } from '@laurence79/express-async-request-handler';\n    import { ParsedQs } from 'qs';\n    import { injectable } from 'tsyringe';\n    import { Validator, ValidationError } from 'express-json-validator-middleware';\n    \n    ".concat(Object.entries(types)
        .map(function (_a) {
        var _b = tslib.__read(_a, 2), name = _b[0], def = _b[1];
        return "export type ".concat(name, " = ").concat(def, ";");
    })
        .join('\n\n'), "\n\n    ").concat(controllers, "\n  ");
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

var PERMITTED_TYPES$1 = ['path', 'query', 'header', 'cookie'];
var convertFormDataToSchemaObject = function (formData) {
    if (!formData || !formData.schema)
        return undefined;
    return tslib.__assign({ type: 'object' }, formData.schema);
};
var fromV3$2 = function (document, log) {
    var _a, _b;
    var definitions = {};
    mapOperations$1(document).forEach(function (_a) {
        var _b;
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses, requestBody = _a.requestBody;
        if (requestBody) {
            var typeName = parametersTypeNameTemplate(operationId, 'body');
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName)));
            var structuredContent = [
                (_b = requestBody.content['application/json']) === null || _b === void 0 ? void 0 : _b.schema,
                convertFormDataToSchemaObject(requestBody.content['multipart/form-data'])
            ]
                .compact()
                .map(convertSchemaToJsonSchema$1);
            if (structuredContent.length > 0) {
                definitions[typeName] =
                    structuredContent.length > 1
                        ? {
                            anyOf: structuredContent
                        }
                        : structuredContent[0];
            }
        }
        var groups = parameters.groupBy(function (p) { return p.in; });
        groups.forEach(function (_a) {
            var key = _a.key, values = _a.values;
            if (!PERMITTED_TYPES$1.includes(key)) {
                throw new Error("Unsupported parameter in ".concat(key));
            }
            var typeName = parametersTypeNameTemplate(operationId, key);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            values.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                var propertyType = (function () {
                    if (!rest.schema) {
                        return { type: 'string' };
                    }
                    if ('$ref' in rest.schema) {
                        return rest.schema;
                    }
                    if (rest.schema.type === 'array') {
                        return {
                            anyOf: [rest.schema, rest.schema.items]
                        };
                    }
                    return rest.schema;
                })();
                schema.properties[name] = propertyType;
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required), false), [name], false);
                }
            });
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName)));
            definitions[typeName] = convertSchemaToJsonSchema$1(schema);
        });
        responses.forEach(function (_a) {
            var _b;
            var statusCode = _a.statusCode, response = _a.response;
            if (!((_b = response.content) === null || _b === void 0 ? void 0 : _b['application/json'])) {
                return;
            }
            var schema = response.content['application/json'].schema;
            if (!schema) {
                return;
            }
            var key = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(key)));
            definitions[key] = convertSchemaToJsonSchema$1(schema);
        });
    });
    Object.keys((_b = (_a = document.components) === null || _a === void 0 ? void 0 : _a.schemas) !== null && _b !== void 0 ? _b : {}).forEach(function (key) {
        var _a, _b;
        var value = (_b = (_a = document.components) === null || _a === void 0 ? void 0 : _a.schemas) === null || _b === void 0 ? void 0 : _b[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(key)));
            definitions[key] = convertSchemaToJsonSchema$1(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: definitions
    };
    return out;
};

var PERMITTED_TYPES = ['body', 'path', 'query', 'header', 'formData'];
var fromV2$2 = function (document, log) {
    var _a;
    var definitions = {};
    mapOperations(document).forEach(function (_a) {
        var operationId = _a.operationId, parameters = _a.parameters, responses = _a.responses;
        var groups = parameters.groupBy(function (p) { return p.in; });
        var keys = groups.map(function (g) { return g.key; });
        if (keys.includes('body') && keys.includes('formData')) {
            throw new Error("Can not have both formData and body parameters");
        }
        groups.forEach(function (_a) {
            var key = _a.key, values = _a.values;
            if (!PERMITTED_TYPES.includes(key)) {
                throw new Error("Unsupported parameter in ".concat(key));
            }
            if (key === 'body') {
                if (values.length > 1) {
                    throw new Error('Only a single body parameter per operation is supported');
                }
                var schema_1 = values[0].schema;
                var typeName_1 = parametersTypeNameTemplate(operationId, 'body');
                log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName_1)));
                definitions[typeName_1] = convertSchemaToJsonSchema(schema_1);
                return;
            }
            var typeName = parametersTypeNameTemplate(operationId, key);
            var schema = {
                type: 'object',
                properties: {},
                required: []
            };
            values.forEach(function (v) {
                // eslint-disable-next-line @typescript-eslint/naming-convention
                v.in; var name = v.name, required = v.required, rest = tslib.__rest(v, ["in", "name", "required"]);
                if (rest.type === 'file') {
                    return;
                }
                var propertyType = (function () {
                    if (rest.type === 'array') {
                        return { anyOf: [rest, rest.items] };
                    }
                    return rest;
                })();
                schema.properties[name] = propertyType;
                if (required) {
                    schema.required = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(schema.required), false), [name], false);
                }
            });
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName)));
            definitions[typeName] = convertSchemaToJsonSchema(schema);
        });
        responses.forEach(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return;
            }
            if (schema.type === 'file') {
                return;
            }
            var key = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(key)));
            definitions[key] = convertSchemaToJsonSchema(schema);
        });
    });
    Object.keys((_a = document.definitions) !== null && _a !== void 0 ? _a : {}).forEach(function (key) {
        var _a;
        var value = (_a = document.definitions) === null || _a === void 0 ? void 0 : _a[key];
        if (value) {
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(key)));
            definitions[key] = convertSchemaToJsonSchema(value);
        }
    });
    var out = {
        $schema: 'http://json-schema.org/draft-07/schema#',
        definitions: definitions
    };
    return out;
};

var generateJsonSchema = function (openApiDocument, log) {
    log === null || log === void 0 ? void 0 : log(progress("Generating JSON schema from Open API Document"));
    var schema = 'swagger' in openApiDocument
        ? fromV2$2(openApiDocument, log)
        : fromV3$2(openApiDocument, log);
    return schema;
};

var fromV2$1 = function (document, log) {
    return mapOperations(document).flatMap(function (_a) {
        var operationId = _a.operationId, responses = _a.responses;
        return responses.compactMap(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var schema = response.schema;
            if (!schema) {
                return undefined;
            }
            if (schema.type !== 'file') {
                return undefined;
            }
            var typeName = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName)));
            return "export type ".concat(typeName, " = Buffer;");
        });
    });
};

var fromV3$1 = function (document, log) {
    return mapOperations$1(document).flatMap(function (_a) {
        var operationId = _a.operationId, responses = _a.responses;
        return responses.compactMap(function (_a) {
            var _b;
            var statusCode = _a.statusCode, response = _a.response;
            if ((_b = response.content) === null || _b === void 0 ? void 0 : _b['application/json']) {
                return undefined;
            }
            var typeName = responseBodyTypeNameTemplate(operationId, statusCode);
            log === null || log === void 0 ? void 0 : log(progress("adding ".concat(typeName)));
            return "export type ".concat(typeName, " = Buffer;");
        });
    });
};

var typeDefForObject = function (objectSchema, recursiveLookup, options) {
    var required = objectSchema.required, properties = objectSchema.properties;
    if (!properties)
        return 'unknown';
    var requiredProperties = Array.isArray(required) ? required : [];
    var propertyList = Object.keys(properties).compactMap(function (propertyName) {
        if (typeof propertyName !== 'string') {
            return undefined;
        }
        var propertyObject = properties[propertyName];
        var isRequired = requiredProperties.includes(propertyName);
        var propertyType = recursiveLookup(propertyObject, options);
        return fieldTemplate(propertyName, isRequired, propertyType, options);
    });
    return "{ ".concat(propertyList.join('; '), " }");
};

var isObjectSchema = function (o) {
    return o.type === 'object' || 'properties' in o;
};

var typeDefForEnum = function (members, onwardLookup, options) {
    return members
        .map(function (e) {
        if (typeof e === 'string') {
            return "\"".concat(e, "\"");
        }
        if (e === null) {
            return 'null';
        }
        if (typeof e === 'object') {
            if (Array.isArray(e)) {
                throw new Error("Don't know what to do with an array-like enum member");
            }
            return onwardLookup(e, options);
        }
        return e.toString();
    })
        .join(' | ');
};

var typeDefForReference = function ($ref) {
    var name = $ref.substring($ref.lastIndexOf('/') + 1);
    return initUpper(safeName(name));
};

var typeDefForSchema = function (schema, options) {
    if (typeof schema === 'boolean') {
        throw new Error("Schema nodes must be an object.");
    }
    if (Array.isArray(schema.type)) {
        throw new Error("Array schema types are not supported (until we understand what they're for).");
    }
    var preliminaryType = (function () {
        if ('$ref' in schema && schema.$ref) {
            return typeDefForReference(schema.$ref);
        }
        var enumProp = schema.enum, allOf = schema.allOf, oneOf = schema.oneOf, anyOf = schema.anyOf;
        if (enumProp) {
            return typeDefForEnum(enumProp, typeDefForSchema, options);
        }
        var base = isObjectSchema(schema)
            ? typeDefForObject(schema, typeDefForSchema, options)
            : null;
        if (allOf) {
            return allOf
                .map(function (m) { return typeDefForSchema(m, options); })
                .concat(base ? [base] : [])
                .join(' & ');
        }
        var unionOf = (oneOf !== null && oneOf !== void 0 ? oneOf : []).concat(anyOf !== null && anyOf !== void 0 ? anyOf : []);
        if (unionOf.length > 0) {
            var union = unionOf.map(function (m) { return typeDefForSchema(m, options); }).join(' | ');
            return base ? "".concat(base, " & (").concat(union, ")") : union;
        }
        if (isObjectSchema(schema)) {
            return base !== null && base !== void 0 ? base : 'unknown';
        }
        if (schema.type === 'boolean' ||
            schema.type === 'null' ||
            schema.type === 'number' ||
            schema.type === 'string') {
            return schema.type;
        }
        if (schema.type === 'integer') {
            return 'number';
        }
        if (schema.type === 'array') {
            var items = schema.items;
            if (!items) {
                throw new Error("Array nodes must have an items property.");
            }
            if (Array.isArray(items)) {
                throw new Error("Array items property must be singular.");
            }
            return "Array<".concat(typeDefForSchema(items, options), ">");
        }
        return 'unknown';
    })();
    if ('nullable' in schema && schema.nullable) {
        return "".concat(preliminaryType, " | null");
    }
    return preliminaryType;
};

var fromJsonSchema = function (jsonSchema, options, log) {
    var definitions = jsonSchema.definitions;
    if (!definitions)
        return [];
    return Object.keys(definitions).compactMap(function (typeName) {
        if (typeof typeName !== 'string')
            return undefined;
        var definition = definitions[typeName];
        log === null || log === void 0 ? void 0 : log(progress(typeName));
        return "export type ".concat(initUpper(safeName(typeName)), " = ").concat(typeDefForSchema(definition, options), ";");
    });
};

var generateServerTypes = function (openApiDocument, jsonSchema, options, log) {
    log === null || log === void 0 ? void 0 : log(progress('Generating typescript types'));
    var types = tslib.__spreadArray(tslib.__spreadArray([], tslib.__read(fromJsonSchema(jsonSchema, options, log)), false), tslib.__read(('swagger' in openApiDocument
        ? fromV2$1(openApiDocument, log)
        : fromV3$1(openApiDocument, log))), false);
    log === null || log === void 0 ? void 0 : log(success());
    return types.join('\n\n');
};

var fromV2 = function (document, log) {
    return mapOperations(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, responses = _a.responses;
        log === null || log === void 0 ? void 0 : log(progress("Adding ".concat(method, " ").concat(path)));
        var bodyArg = parameters.first(function (p) { return p.in === 'body' || (p.in === 'formData' && p.type !== 'file'); });
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var headersType = headerArgs.any()
            ? parametersTypeNameTemplate(operationId, 'header')
            : 'unknown';
        var pathParamsType = pathArgs.any()
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var queryParamsType = queryArgs.any()
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var requestBodyType = bodyArg
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var responseStubType = responses
            .map(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var statusCodeType = statusCode === 'default' ? 'number' : statusCode;
            var responseBodyType = (function () {
                var schema = response.schema;
                if (!schema) {
                    return 'undefined';
                }
                return responseBodyTypeNameTemplate(operationId, statusCode);
            })();
            return "ResponseStub<".concat(statusCodeType, ", ").concat(responseBodyType, ">");
        })
            .distinct()
            .join(' | ') || 'ResponseStub';
        var methodStubType = "MethodStub<\n        ".concat(headersType, ",\n        ").concat(pathParamsType, ",\n        ").concat(queryParamsType, ",\n        ").concat(requestBodyType, ",\n        ").concat(responseStubType, "\n      >");
        var requestHandlerPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":".concat(x.substr(1, x.length - 2)); });
        return {
            operationId: operationId,
            requestHandlerPath: requestHandlerPath,
            method: method,
            methodStubType: methodStubType
        };
    });
};

var fromV3 = function (document, log) {
    return mapOperations$1(document).map(function (_a) {
        var method = _a.method, path = _a.path, parameters = _a.parameters, operationId = _a.operationId, responses = _a.responses, requestBody = _a.requestBody;
        log === null || log === void 0 ? void 0 : log(progress("Adding ".concat(method, " ").concat(path)));
        var pathArgs = parameters.filter(function (p) { return p.in === 'path'; });
        var queryArgs = parameters.filter(function (p) { return p.in === 'query'; });
        var headerArgs = parameters.filter(function (p) { return p.in === 'header'; });
        var headersType = headerArgs.any()
            ? parametersTypeNameTemplate(operationId, 'header')
            : 'unknown';
        var pathParamsType = pathArgs.any()
            ? parametersTypeNameTemplate(operationId, 'path')
            : 'unknown';
        var queryParamsType = queryArgs.any()
            ? parametersTypeNameTemplate(operationId, 'query')
            : 'unknown';
        var requestBodyType = requestBody
            ? parametersTypeNameTemplate(operationId, 'body')
            : 'unknown';
        var responseStubType = responses
            .map(function (_a) {
            var statusCode = _a.statusCode, response = _a.response;
            var statusCodeType = statusCode === 'default' ? 'number' : statusCode;
            var responseBodyType = (function () {
                if (!response.content ||
                    Object.keys(response.content).length === 0) {
                    return 'undefined';
                }
                return responseBodyTypeNameTemplate(operationId, statusCode);
            })();
            return "ResponseStub<".concat(statusCodeType, ", ").concat(responseBodyType, ">");
        })
            .distinct()
            .join(' | ') || 'unknown';
        var methodStubType = "MethodStub<\n        ".concat(headersType, ",\n        ").concat(pathParamsType, ",\n        ").concat(queryParamsType, ",\n        ").concat(requestBodyType, ",\n        ").concat(responseStubType, "\n      >");
        var requestHandlerPath = path.replace(/\{(?:.*?)\}/g, function (x) { return ":".concat(x.substr(1, x.length - 2)); });
        return {
            operationId: operationId,
            requestHandlerPath: requestHandlerPath,
            method: method,
            methodStubType: methodStubType
        };
    });
};

var generateStubs = function (_a) {
    var logger = _a.logger, openApiDocument = _a.openApiDocument, serviceName = _a.serviceName, nonRequiredType = _a.nonRequiredType;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Generating server stubs');
    var jsonSchema = generateJsonSchema(openApiDocument, log);
    var types = generateServerTypes(openApiDocument, jsonSchema, { nonRequiredType: nonRequiredType }, log);
    var data = 'swagger' in openApiDocument
        ? fromV2(openApiDocument, log)
        : fromV3(openApiDocument, log);
    var code = "\n    import { RequestHandler } from '@laurence79/express-async-request-handler';\n    import { Router } from 'express';\n\n    ".concat(types, "\n      \n    type RequestLog<THeaders, TParams, TQuery, TBody> = {\n      headers: THeaders;\n      params: TParams;\n      query: TQuery;\n      body: TBody;\n    };\n\n    type ResponseStub<TStatus extends number = number, TBody = unknown> = {\n      statusCode: TStatus;\n      body: TBody;\n    };\n\n    type MethodStub<\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TResponseStub extends ResponseStub\n    > = {\n      requestHandler: RequestHandler<\n        TParams,\n        TResponseStub['body'],\n        TRequestBody,\n        TQuery,\n        Record<string, unknown>,\n        TResponseStub['statusCode']\n      >;\n      respondWith(data: TResponseStub | TResponseStub[]): void;\n      callLog(): RequestLog<THeaders, TParams, TQuery, TRequestBody>[];\n    };\n\n    const methodStub = <\n      THeaders,\n      TParams,\n      TQuery,\n      TRequestBody,\n      TResponseStub extends ResponseStub\n    >(): MethodStub<THeaders, TParams, TQuery, TRequestBody, TResponseStub> => {\n      const log: RequestLog<THeaders, TParams, TQuery, TRequestBody>[] = [];\n\n      let responseStub: TResponseStub | TResponseStub[] | null = null;\n\n      return {\n        requestHandler: (req, res) => {\n          const { headers, params, query, body } = req;\n\n          log.push({\n            headers: (headers as unknown) as THeaders,\n            params,\n            query,\n            body\n          });\n\n          if (!responseStub) {\n            throw new Error('Method not stubbed');\n          }\n\n          if (Array.isArray(responseStub)) {\n            const stub = responseStub.shift();\n\n            if (!stub) {\n              throw new Error('No more stubbed responses left');\n            }\n\n            return res.status(stub.statusCode).send(stub.body);\n          }\n\n          return res.status(responseStub.statusCode).send(responseStub.body);\n        },\n\n        respondWith(data) {\n          responseStub = data;\n        },\n\n        callLog() {\n          return log;\n        }\n      };\n    };\n\n    type MethodStubs = {\n      ").concat(data
        .map(function (_a) {
        var operationId = _a.operationId, methodStubType = _a.methodStubType;
        return "readonly ".concat(safeName(operationId), ": ").concat(methodStubType, ";");
    })
        .join('\n'), "\n    };\n\n    export type ").concat(serviceName, "ServiceStub = {\n      readonly middleware: RequestHandler;\n      readonly reset: () => void;\n      readonly stubs: () => MethodStubs;\n    };\n\n    export const ").concat(initLower(serviceName), "ServiceStub \n    = (): ").concat(serviceName, "ServiceStub => {\n      let methodStubs: MethodStubs;\n      let currentRouter: RequestHandler;\n\n      const reset = () => {\n        methodStubs = {\n          ").concat(data
        .map(function (_a) {
        var operationId = _a.operationId;
        return "".concat(safeName(operationId), ": methodStub()");
    })
        .join(',\n'), "\n        };\n\n        currentRouter = Router()\n          ").concat(data
        .map(function (_a) {
        var method = _a.method, operationId = _a.operationId, requestHandlerPath = _a.requestHandlerPath;
        return ".".concat(method, "('").concat(requestHandlerPath, "', methodStubs.").concat(safeName(operationId), ".requestHandler)");
    })
        .join('\n'), "\n      };\n\n      reset();\n\n      return {\n        reset,\n        middleware: (...args: Parameters<RequestHandler>) => currentRouter(...args),\n        stubs: () => methodStubs\n      };\n    };\n  ");
    log === null || log === void 0 ? void 0 : log(success());
    return code;
};

/* eslint-disable @typescript-eslint/restrict-template-expressions */
function assertNever(x) {
    throw new Error("Unexpected object: ".concat(x));
}

var fileExists = util.promisify(fs__default["default"].exists);
var readFile = util.promisify(fs__default["default"].readFile);
var loadFromFile = function (source, logger) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    var log, extension, data, doc;
    return tslib.__generator(this, function (_a) {
        switch (_a.label) {
            case 0:
                log = logger === null || logger === void 0 ? void 0 : logger.create("Loading OpenAPI file ".concat(source));
                return [4 /*yield*/, fileExists(source)];
            case 1:
                if (!(_a.sent())) {
                    throw new Error("File ".concat(source, " does not exist"));
                }
                extension = path__default["default"].extname(source);
                return [4 /*yield*/, readFile(source, 'utf-8')];
            case 2:
                data = _a.sent();
                doc = (function () {
                    if (['.yaml', '.yml'].includes(extension)) {
                        return yaml__default["default"].parse(data);
                    }
                    if (extension === '.json') {
                        return JSON.parse(data);
                    }
                    throw new Error("Unsupported extension ".concat(extension));
                })();
                log === null || log === void 0 ? void 0 : log(success());
                return [2 /*return*/, doc];
        }
    });
}); };
var isReference = function (node) {
    return (typeof node === 'object' &&
        node !== null &&
        '$ref' in node &&
        typeof node.$ref === 'string');
};
var isWalkable = function (node) {
    return typeof node === 'object' && node !== null;
};
var getAbsolutePath = function (refBasePath, referencePath) {
    if (path__default["default"].isAbsolute(referencePath)) {
        return referencePath;
    }
    return path__default["default"].join(path__default["default"].dirname(refBasePath), referencePath);
};
var inferTitle = function (referencePath, pointer) {
    if (pointer) {
        return pointer.substring(pointer.lastIndexOf('/') + 1);
    }
    if (referencePath) {
        return path__default["default"].parse(referencePath).name;
    }
    return null;
};
var dereference = function (refBasePath, reference, files, logger) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    var $ref, _a, referencePath, pointer, file, refDoc, node;
    return tslib.__generator(this, function (_b) {
        switch (_b.label) {
            case 0:
                $ref = reference.$ref;
                _a = tslib.__read($ref.split('#'), 2), referencePath = _a[0], pointer = _a[1];
                file = referencePath
                    ? getAbsolutePath(refBasePath, referencePath)
                    : refBasePath;
                return [4 /*yield*/, (function () { return tslib.__awaiter(void 0, void 0, void 0, function () {
                        var cachedDoc, doc;
                        return tslib.__generator(this, function (_a) {
                            switch (_a.label) {
                                case 0:
                                    cachedDoc = files.get(file);
                                    if (!!cachedDoc) return [3 /*break*/, 3];
                                    return [4 /*yield*/, loadFromFile(file)];
                                case 1:
                                    doc = _a.sent();
                                    if (!isWalkable(doc)) {
                                        throw new Error("Can not walk ".concat(file));
                                    }
                                    files.set(file, doc);
                                    // eslint-disable-next-line @typescript-eslint/no-use-before-define
                                    return [4 /*yield*/, walk(doc, file, files)];
                                case 2:
                                    // eslint-disable-next-line @typescript-eslint/no-use-before-define
                                    _a.sent();
                                    return [2 /*return*/, doc];
                                case 3: return [2 /*return*/, cachedDoc];
                            }
                        });
                    }); })()];
            case 1:
                refDoc = _b.sent();
                node = pointer ? jsonPointer__default["default"].get(refDoc, pointer) : refDoc;
                if (isReference(node)) {
                    return [2 /*return*/, dereference(file, node, files)];
                }
                if (typeof node === 'object' && node && !Array.isArray(node) && !node.title) {
                    node.title = inferTitle(referencePath, pointer);
                }
                return [2 /*return*/, node];
        }
    });
}); };
var walkArray = function (node, refBasePath, files, logger) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    return tslib.__generator(this, function (_a) {
        switch (_a.label) {
            case 0: return [4 /*yield*/, node.forEachAsync(function (item, index) { return tslib.__awaiter(void 0, void 0, void 0, function () {
                    var _a, _b;
                    return tslib.__generator(this, function (_c) {
                        switch (_c.label) {
                            case 0:
                                if (!isReference(item)) return [3 /*break*/, 2];
                                _a = node;
                                _b = index;
                                return [4 /*yield*/, dereference(refBasePath, item, files)];
                            case 1:
                                _a[_b] = _c.sent();
                                return [2 /*return*/];
                            case 2:
                                if (!isWalkable(item)) return [3 /*break*/, 4];
                                // eslint-disable-next-line @typescript-eslint/no-use-before-define
                                return [4 /*yield*/, walk(item, refBasePath, files)];
                            case 3:
                                // eslint-disable-next-line @typescript-eslint/no-use-before-define
                                _c.sent();
                                _c.label = 4;
                            case 4: return [2 /*return*/];
                        }
                    });
                }); })];
            case 1:
                _a.sent();
                return [2 /*return*/];
        }
    });
}); };
var walkMap = function (node, refBasePath, files, logger) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    return tslib.__generator(this, function (_a) {
        switch (_a.label) {
            case 0: return [4 /*yield*/, Object.entries(node).forEachAsync(function (_a) {
                    var _b = tslib.__read(_a, 2), key = _b[0], item = _b[1];
                    return tslib.__awaiter(void 0, void 0, void 0, function () {
                        var _c, _d;
                        return tslib.__generator(this, function (_e) {
                            switch (_e.label) {
                                case 0:
                                    if (!isReference(item)) return [3 /*break*/, 2];
                                    _c = node;
                                    _d = key;
                                    return [4 /*yield*/, dereference(refBasePath, item, files)];
                                case 1:
                                    _c[_d] = _e.sent();
                                    return [2 /*return*/];
                                case 2:
                                    if (!isWalkable(item)) return [3 /*break*/, 4];
                                    // eslint-disable-next-line @typescript-eslint/no-use-before-define
                                    return [4 /*yield*/, walk(item, refBasePath, files)];
                                case 3:
                                    // eslint-disable-next-line @typescript-eslint/no-use-before-define
                                    _e.sent();
                                    _e.label = 4;
                                case 4: return [2 /*return*/];
                            }
                        });
                    });
                })];
            case 1:
                _a.sent();
                return [2 /*return*/];
        }
    });
}); };
var walk = function (node, refBasePath, files, logger) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    return tslib.__generator(this, function (_a) {
        switch (_a.label) {
            case 0:
                if (!Array.isArray(node)) return [3 /*break*/, 2];
                return [4 /*yield*/, walkArray(node, refBasePath, files)];
            case 1:
                _a.sent();
                return [2 /*return*/];
            case 2: return [4 /*yield*/, walkMap(node, refBasePath, files)];
            case 3:
                _a.sent();
                return [2 /*return*/];
        }
    });
}); };
var load = function (source, logger) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    var data, files;
    return tslib.__generator(this, function (_a) {
        switch (_a.label) {
            case 0: return [4 /*yield*/, loadFromFile(source, logger)];
            case 1:
                data = _a.sent();
                if (!isWalkable(data)) {
                    throw new Error("Can not walk ".concat(source));
                }
                files = new Map();
                files.set(source, data);
                return [4 /*yield*/, walk(data, source, files)];
            case 2:
                _a.sent();
                return [2 /*return*/, data];
        }
    });
}); };

var prettierStyle = {
    arrowParens: 'avoid',
    semi: true,
    trailingComma: 'none',
    singleQuote: true,
    printWidth: 80
};
var writeFile = function (_a) {
    var content = _a.content, logger = _a.logger, filename = _a.filename;
    var log = logger === null || logger === void 0 ? void 0 : logger.create('Writing file');
    log === null || log === void 0 ? void 0 : log(progress('Creating output directory'));
    fs__default["default"].mkdirSync(path__default["default"].dirname(filename), { recursive: true });
    log === null || log === void 0 ? void 0 : log(progress('Formatting content'));
    var extension = path__default["default"].extname(filename);
    var parser = (function () {
        switch (extension) {
            case '.ts':
                return 'typescript';
            case '.json':
                return 'json';
            default:
                return undefined;
        }
    })();
    var formattedContent = (function () {
        try {
            return prettier__default["default"].format(content, tslib.__assign(tslib.__assign({}, prettierStyle), { parser: parser }));
        }
        catch (e) {
            if (e instanceof Error) {
                log === null || log === void 0 ? void 0 : log(warn(e.message));
            }
            return content;
        }
    })();
    fs__default["default"].writeFileSync(filename, formattedContent);
    log === null || log === void 0 ? void 0 : log(success());
};

var defaultOptions = {
    output: 'CLIENT'
};
var generateCode = function (inputOptions) { return tslib.__awaiter(void 0, void 0, void 0, function () {
    var options, openApiDocument, serviceName, filename, renderDeps, method;
    var _a;
    return tslib.__generator(this, function (_b) {
        switch (_b.label) {
            case 0:
                options = tslib.__assign(tslib.__assign({}, defaultOptions), inputOptions);
                return [4 /*yield*/, load(options.openApiDocumentFilenameOrUrl, options.logger)];
            case 1:
                openApiDocument = (_b.sent());
                serviceName = (_a = inputOptions.serviceName) !== null && _a !== void 0 ? _a : serviceNameTemplate(openApiDocument.info.title);
                filename = (function () {
                    var outputTypeName = initUpper(options.output.toLowerCase().replace('-', ' ')).replace(' ', '');
                    var file = "".concat(serviceName).concat(outputTypeName, ".generated.ts");
                    if (!options.outputFilename) {
                        return file;
                    }
                    if (path__default["default"].extname(options.outputFilename)) {
                        return options.outputFilename;
                    }
                    return path__default["default"].join(options.outputFilename, file);
                })();
                renderDeps = tslib.__assign(tslib.__assign({}, options), { openApiDocument: openApiDocument, serviceName: serviceName, filename: filename });
                method = (function () {
                    switch (options.output) {
                        case 'CLIENT':
                            return generateClient;
                        case 'SERVER':
                            return generateServerClasses;
                        case 'STUBS':
                            return generateStubs;
                        default:
                            return assertNever(options.output);
                    }
                })();
                writeFile(tslib.__assign(tslib.__assign({}, renderDeps), { content: method(renderDeps) }));
                return [2 /*return*/];
        }
    });
}); };

exports.generateCode = generateCode;
exports.processLogger = processLogger;
